// Itool.java: a tool for  browsing in the Interface Repository and calling
//             Methods of Corba-Objects.
// 
// last modified: 25.06.97 by Tim Paehler
// 

import java.util.*;
import java.io.*;
import IE.Iona.Orbix2.*;
import IE.Iona.Orbix2.CORBA.*;

public class Itool {

// We are an application, therefore all our methods and variables are static


// 'global' variables 
static ORB orb = ORB.IT_create();
static _RepositoryRef repRef;
static int current_level;
static _ContainerRef[] container = new _ContainerRef[100];
static String prompt;


  /***********************************************************
 /   cd: change the current container, called by usercommand /
/  args: name - the name of the object to evaluated         /
***********************************************************/
private static void cd(String name){

  // check whether ascension or descension in the tree was specified

  if( name.equals("/")){
    current_level = 0;
    return;
  }
 
  if( name.equals("..")){
    if(current_level != 0)
      current_level--;
    return ;
  }
 
  // now we try to descend into the specified container
  
  try {
    ContainedSeq found = container[current_level].lookup_name(name,1,"all",true);
    if(found.length>0)
      container[current_level+1] = Container._narrow(found.buffer[0]);
    else {
      System.out.println(name+": No such container"); 
      return; }
  } 
  catch (SystemException e){ 
    System.out.println(name+": No such container"); return; }
  current_level++;
}
  
   /****************************************************************
  / evaluate_object: format output for 'ls'-command                /
 /             args: value and name to be evaluated               /
/   returns a description to be displayed                        /
 ***************************************************************/
private static String evaluate_irobject(Any value,String name){
  ModuleDescription mod_desc = new ModuleDescription();
  InterfaceDescription if_desc = new InterfaceDescription();
  OperationDescription op_desc = new OperationDescription();
  ParameterDescription parm_desc = new ParameterDescription();
  AttributeDescription attr_desc = new AttributeDescription();
  ExceptionDescription exc_desc = new ExceptionDescription();
  TypeDescription type_desc = new TypeDescription();
  ConstantDescription const_desc = new ConstantDescription();
  String parstring = "";
  int i;
  String test;
  Hashtable typecodes = new Hashtable(15);
  Hashtable modes = new Hashtable(3);

  // build translation tables for constants defined in Package CORBA
  modes.put("0","in");
  modes.put("1","out");
  modes.put("2","inout");
  typecodes.put("v","void");
  typecodes.put("b","boolean");
  typecodes.put("s","short");
  typecodes.put("l","long");
  typecodes.put("c","char");
  typecodes.put("0","string");
  typecodes.put("T","TypeCode");
  typecodes.put("a","Any");

  // check and return the type of description in value
  try{
    if(name.equals("ModuleDescription")){
      value.extract(mod_desc);
      return mod_desc.name;
    }

    if(name.equals("InterfaceDescription")){
      value.extract(if_desc);
      return if_desc.name;
    }

    // if we are dealing with an operation, output must be additionally 
    // formatted to show parameters and their type/mode

    if(name.equals("OperationDescription")){
      value.extract(op_desc);

      // get and format parameters to 
      // 'result operation( in char blubb, out int bla..)'
      for(i=op_desc.parameter.length-1;i>=0;i--) {
	parstring =
	  modes.get(new Integer(op_desc.parameter.buffer[i].mode).toString())
	  +" "+(typecodes.get(test = op_desc.parameter.buffer[i].type.toString()) == "null" ? test : typecodes.get(test))+" "
          +op_desc.parameter.buffer[i].name+", "+parstring;
      }
      if(op_desc.parameter.length>0) // chop last ','
	parstring=parstring.substring(0,parstring.length()-2);
      return (typecodes.get(test = op_desc.result.toString()) == null ? test : typecodes.get(test))+" "+op_desc.name+"("+parstring+")";
    } 
    
    if(name.equals("ParameterDescription")){
      value.extract(parm_desc);
      return parm_desc.name;
    }

    if(name.equals("AttributeDescription")){
      value.extract(attr_desc);
      return (String) modes.get(new Integer(attr_desc.mode).toString()) +" "
             + (typecodes.get(test = attr_desc.type.toString()) == null ? test : typecodes.get(test)) 
	     + " "+attr_desc.name;
    }

    if(name.equals("ExceptionDescription")){
      value.extract(exc_desc);
      return exc_desc.name;
    }

    if(name.equals("TypeDescription")){
      value.extract(type_desc);
      return type_desc.name;
    }

  } 
  catch(SystemException e){}
  return "Unknown: "+name+"!"; 

}

  /**************************************************
 /   ls: list the content of current container     /
/  args: arg - 
 *************************************************/

private static void ls(String arg){
  int i=0;
  ContainedSeq contained;
  try {
    contained = container[current_level].contents(arg,true);
    for(i=contained.length-1;i>=0;i--){
      String name = contained.buffer[i].describe().name;
      System.out.print(name.substring(0,(name.length()-11))+": ");
      System.out.println(evaluate_irobject(contained.buffer[i].describe().value,name));
    }
  } catch (SystemException e){ 
    System.out.println("Error:"+e); 
    return;
  }

}


  /***************************************************************************
 /   callByObjString: resolve a remote method-invocation by using the DII    /
/                arg: str - the object-reference + method-invocation syntax /
***************************************************************************/

private static void callByObjString(String str){
  
  Any value;
  Any[] args = new Any[10];
  _ObjectRef objRef = null;
  int i = 0;
  Object[] holders = new Object[20];
  int count = 0;

  // parse the String str, retrieve object-reference, operation-name and 
  // parameters

  String objString = str.substring(0,str.indexOf(' '));
  String method = str.substring(str.indexOf(' ')+1,str.length());
  String operation = str.substring(str.indexOf(' ')+1,str.indexOf("("));
  str = str.substring(str.indexOf("(")+1,str.lastIndexOf(")"));
  StringTokenizer strtok = new StringTokenizer(str,",");
  String[] parameters = new String[strtok.countTokens()];
  while(strtok.hasMoreTokens()){
    parameters[i++] = strtok.nextToken();
  }

  try { 
    String method_line = null;
    System.out.println("try string_to_object ...");
    objRef = _CORBA.Orbix.string_to_object(objString);
    
    // get the operation of the corresponding interface from the IR

    _InterfaceDefRef ifRef = objRef._get_interface();    
    ContainedSeq found = (Container._narrow(ifRef)).lookup_name(operation,1,"all",true);
    if(found.length>0){
      _OperationDefRef opdefRef = OperationDef._narrow(found.buffer[0]);
            
      // get the types of required parameters from IR-query output

      method_line = evaluate_irobject(Contained._narrow(opdefRef).describe().value,Contained._narrow(opdefRef).describe().name);
    }
    else {

      // No Entry in IR - User must interactively specify types of arguments 

      System.out.print("No Entry in IR for "+operation+". Enter method syntax:<type> <method>(<mode> <type> <arg1>, <mode> <type> <arg2>)\n$");
      System.out.flush();
      DataInputStream is = new DataInputStream(System.in);
      try{
	method_line=is.readLine();
      }catch(IOException e){};
    }
    System.out.println("Method:\""+method_line+"\"");
    
    System.out.println("try request on Object "+objRef+"...");
  
    Request request = new Request(objRef,operation);
    
    // insert the parameters, type- and mode-information retrieved from 
    // method_line 

    // parse method_line
    strtok = new StringTokenizer(method_line.substring(method_line.indexOf("(")+1,method_line.lastIndexOf(")")),",");
    String[] required_parameters = new String[strtok.countTokens()];
    i=0;
    while(strtok.hasMoreTokens())
      required_parameters[i++]=strtok.nextToken();

    // now loop through parameters and insert them into the request-NVList
    for(i=0;i<parameters.length;i++)
      {
	// get type- and mode-information from required_parameters
	required_parameters[i].trim();
	StringTokenizer strtok2 = new StringTokenizer(required_parameters[i]," ");
	String flags = strtok2.nextToken();
	String type = strtok2.nextToken();
	boolean holder = false; 
	if(flags.equals("in"))
	   request.setArgFlags (_CORBA.ARG_IN);
	if(flags.equals("out")){
	   request.setArgFlags (_CORBA.ARG_OUT);
	   holder = true;}
	if(flags.equals("inout")){
	   request.setArgFlags (_CORBA.ARG_INOUT);
	   holder = true;}
	// CORBA-compliant Parameter-Insertion doesn't work yet, we have to 
        // use the Orbix way
	if(type.equals("short"))
	  if(holder){
	    holders[count] = new ShortHolder((short)new Integer(parameters[i]).intValue());
	    request.insertShort((ShortHolder)holders[count++]);}
	  else
	    request.insertShort((short)(new Integer(parameters[i]).intValue()));
	if(type.equals("long"))
	  if(holder){
	    holders[count] = new IntHolder(new Integer(parameters[i]).intValue());
	    request.insertLong((IntHolder)holders[count++]);}
	  else
	    request.insertLong(new Integer(parameters[i]).intValue());
	if(type.equals("float"))
	  if(holder){
	    holders[count] = new FloatHolder(new Float(parameters[i]).floatValue());
	    request.insertFloat((FloatHolder)holders[count++]);}
	  else
	    request.insertFloat(new Float(parameters[i]).floatValue());
	if(type.equals("double"))
	  if(holder){
	    holders[count] = new DoubleHolder(new Double(parameters[i]).doubleValue());
	    request.insertDouble((DoubleHolder)holders[count++]);}
	  else
	    request.insertDouble(new Double(parameters[i]).doubleValue());
	if(type.equals("char"))
	  if(holder){
	    holders[count] = new CharHolder(new String(parameters[i]).charAt(0));
	    request.insertChar((CharHolder)holders[count++]);}	
	  else
	    request.insertChar((parameters[i]).charAt(0));	
	if(type.equals("string"))
	  if(holder){
	    holders[count] = new StringHolder(new String(parameters[i]));
	    request.insertString((StringHolder)holders[count++]);}
	  else
	    request.insertString((String)(parameters[i]));
	if(type.equals("boolean"))
	  if(holder){
	    holders[count] = new BooleanHolder(((parameters[i]).equals("true") ? true : false));
	    request.insertBoolean((BooleanHolder)holders[count++]);}
	  else
	    request.insertBoolean(((parameters[i]).equals("true") ? true : false));
      }

    // Now send the request to the server and hope it returns safely ;-)

    request.invoke();
    
    // Show result and out/inout-parameters of method-invocation

    String result = method_line.substring(0,method_line.indexOf(" "));

    System.out.print("Result is: "+"("+result+") ");
	if(result.equals("short"))
	  System.out.println(request.extractShort());
	if(result.equals("long"))
	  System.out.println(request.extractLong());
	if(result.equals("float"))
	  System.out.println(request.extractFloat());
	if(result.equals("double"))
	  System.out.println(request.extractDouble());
	if(result.equals("char"))
	  System.out.println(request.extractChar());
	if(result.equals("string"))
	  System.out.println(request.extractString());
	if(result.equals("boolean"))
	  System.out.println(request.extractBoolean());
	System.out.println();
	request.extractOutParams ();
	for(i=0;i<count;i++){
	  if(holders[i] instanceof ShortHolder)
	     System.out.println("out/inout-Value: (short) "+((ShortHolder) holders[i]).value);
	  if(holders[i] instanceof IntHolder)
	     System.out.println("out/inout-Value: (int) "+((IntHolder) holders[i]).value);
	  if(holders[i] instanceof FloatHolder)
	     System.out.println("out/inout-Value: (float) "+((FloatHolder) holders[i]).value);
	  if(holders[i] instanceof DoubleHolder)
	     System.out.println("out/inout-Value: (double) "+((DoubleHolder) holders[i]).value);
	  if(holders[i] instanceof CharHolder)
	     System.out.println("out/inout-Value: (char) "+((CharHolder) holders[i]).value);
	  if(holders[i] instanceof StringHolder)
	     System.out.println("out/inout-Value: (String) "+((StringHolder) holders[i]).value);
	  if(holders[i] instanceof BooleanHolder)
	     System.out.println("out/inout-Value: (boolean) "+((BooleanHolder) holders[i]).value);
	}
  } 
  catch (IE.Iona.Orbix2.CORBA.CORBAException e){ System.err.println("call Error:"+e);}
 
}


  /**********************************************
 /   main-function: parse and execute commands  /
/    no arguments taken                        /
 *********************************************/

public static void main(String[] args){

  _ContainedRef ctd;
  // use the Orbix Protocol and the local orbixd
  _CORBA.IT_BIND_USING_IIOP = false; // IMPORTANT !!!
  // Use stdin as input for command-parsing
  DataInputStream is = new DataInputStream(System.in);
  int i;

  // _bind Repository and display its Object-Reference
  try {
    String id;
    repRef = Repository._bind(":IR","deeppurple");
    container[ current_level=0 ] = Container._narrow(repRef);
    System.out.println(repRef);
  } catch (SystemException e){ System.err.println("My Error: "+e);}
    System.out.println("Possible commands are:\tls [IRObjecttype]\n\t\t\tcd </|..|<container>>\n\t\t\tcall <objectid> method(parameter1,parameter2...)\n\t\t\tquit");
    
    // parse commands until 'quit' is entered
    while(true) {
      String input = null;
      prompt ="/";
      try{
      for(i=1;i<=current_level;i++)
	prompt += "/"+evaluate_irobject((ctd=Contained._narrow(container[i])).describe().value,ctd.describe().name);
      } catch (SystemException e){ System.err.println("My Error: "+e);}
      System.out.print(prompt+"$");
      System.out.flush();
      try{input = is.readLine();}catch(IOException e){}
      if(input.equals("quit") || input.equals("exit") || input.equals("q"))
	System.exit(0);
      else
	if(input.startsWith("cd"))
	  cd(input.substring(3));
	else
	  if(input.startsWith("ls"))
	    if(input.length()>2)
	      ls(input.substring(3));
	    else
	      ls("all");
	  else
	    if(input.startsWith("call"))
	      callByObjString(input.substring(5));
	    else
	      System.out.println("Invalid command: "+input);
    } 
	
}

}
